import logging
import re
from pathlib import Path
from typing import List, Union

import flair
from flair.data import Corpus, FlairDataset, Sentence, Token
from flair.datasets.base import find_train_dev_test_files
from flair.file_utils import cached_path

log = logging.getLogger("flair")


class UniversalDependenciesCorpus(Corpus):
    def __init__(
        self,
        data_folder: Union[str, Path],
        train_file=None,
        test_file=None,
        dev_file=None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):
        """
        Instantiates a Corpus from CoNLL-U column-formatted task data such as the UD corpora

        :param data_folder: base folder with the task data
        :param train_file: the name of the train file
        :param test_file: the name of the test file
        :param dev_file: the name of the dev file, if None, dev data is sampled from train
        :param in_memory: If set to True, keeps full dataset in memory, otherwise does disk reads
        :param split_multiwords: If set to True, multiwords are split (default), otherwise kept as single tokens
        :return: a Corpus with annotated train, dev and test data
        """

        # find train, dev and test files if not specified
        dev_file, test_file, train_file = find_train_dev_test_files(data_folder, dev_file, test_file, train_file)

        # get train data
        train = UniversalDependenciesDataset(train_file, in_memory=in_memory, split_multiwords=split_multiwords)

        # get test data
        test = (
            UniversalDependenciesDataset(test_file, in_memory=in_memory, split_multiwords=split_multiwords)
            if test_file is not None
            else None
        )

        # get dev data
        dev = (
            UniversalDependenciesDataset(dev_file, in_memory=in_memory, split_multiwords=split_multiwords)
            if dev_file is not None
            else None
        )

        super(UniversalDependenciesCorpus, self).__init__(train, dev, test, name=str(data_folder))


class UniversalDependenciesDataset(FlairDataset):
    def __init__(
        self,
        path_to_conll_file: Union[str, Path],
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):
        """
        Instantiates a column dataset in CoNLL-U format.

        :param path_to_conll_file: Path to the CoNLL-U formatted file
        :param in_memory: If set to True, keeps full dataset in memory, otherwise does disk reads
        """
        path_to_conll_file = Path(path_to_conll_file)
        assert path_to_conll_file.exists()

        self.in_memory: bool = in_memory
        self.split_multiwords: bool = split_multiwords

        self.path_to_conll_file = path_to_conll_file
        self.total_sentence_count: int = 0

        with open(str(self.path_to_conll_file), encoding="utf-8") as file:

            # option 1: read only sentence boundaries as offset positions
            if not self.in_memory:
                self.indices: List[int] = []

                line = file.readline()
                position = 0
                while line:
                    line = line.strip()
                    if line == "":
                        self.indices.append(position)
                        position = file.tell()
                    line = file.readline()

                self.total_sentence_count = len(self.indices)

            # option 2: keep everything in memory
            if self.in_memory:
                self.sentences: List[Sentence] = []

                while True:
                    sentence = self._read_next_sentence(file)
                    if not sentence:
                        break
                    self.sentences.append(sentence)

                self.total_sentence_count = len(self.sentences)

    def is_in_memory(self) -> bool:
        return self.in_memory

    def __len__(self):
        return self.total_sentence_count

    def __getitem__(self, index: int = 0) -> Sentence:

        # if in memory, retrieve parsed sentence
        if self.in_memory:
            sentence = self.sentences[index]

        # else skip to position in file where sentence begins
        else:
            with open(str(self.path_to_conll_file), encoding="utf-8") as file:
                file.seek(self.indices[index])
                sentence = self._read_next_sentence(file)

        return sentence

    def _read_next_sentence(self, file):
        line = file.readline()
        sentence: Sentence = Sentence([])

        # current token ID
        token_idx = 0

        # handling for the awful UD multiword format
        current_multiword_text = ""
        current_multiword_sequence = ""
        current_multiword_first_token = 0
        current_multiword_last_token = 0

        while line:
            line = line.strip()
            fields: List[str] = re.split("\t+", line)

            # end of sentence
            if line == "":
                if len(sentence) > 0:
                    break

            # comments
            elif line.startswith("#"):
                line = file.readline()
                continue

            # ellipsis
            elif "." in fields[0]:
                line = file.readline()
                continue

            # if token is a multi-word
            elif "-" in fields[0]:
                line = file.readline()

                current_multiword_first_token = int(fields[0].split("-")[0])
                current_multiword_last_token = int(fields[0].split("-")[1])
                current_multiword_text = fields[1]
                current_multiword_sequence = ""

                if self.split_multiwords:
                    continue
                else:
                    token = Token(fields[1])
                    token.add_label("lemma", str(fields[2]))
                    if len(fields) > 9 and "SpaceAfter=No" in fields[9]:
                        token.whitespace_after = 0
                    sentence.add_token(token)
                    token_idx += 1

            # normal single-word tokens
            else:

                # if we don't split multiwords, skip over component words
                if not self.split_multiwords and token_idx < current_multiword_last_token:
                    token_idx += 1
                    line = file.readline()
                    continue

                # add token
                token = Token(fields[1], head_id=int(fields[6]))
                token.add_label("lemma", str(fields[2]))
                token.add_label("upos", str(fields[3]))
                token.add_label("pos", str(fields[4]))
                token.add_label("dependency", str(fields[7]))

                if len(fields) > 9 and "SpaceAfter=No" in fields[9]:
                    token.whitespace_after = 0

                # add morphological tags
                for morph in str(fields[5]).split("|"):
                    if "=" not in morph:
                        continue
                    token.add_label(morph.split("=")[0].lower(), morph.split("=")[1])

                if len(fields) > 10 and str(fields[10]) == "Y":
                    token.add_label("frame", str(fields[11]))

                token_idx += 1

                # derive whitespace logic for multiwords
                if token_idx <= current_multiword_last_token:
                    current_multiword_sequence += token.text

                # print(token)
                # print(current_multiword_last_token)
                # print(current_multiword_first_token)
                # if multi-word equals component tokens, there should be no whitespace
                if token_idx == current_multiword_last_token and current_multiword_sequence == current_multiword_text:
                    # go through all tokens in subword and set whitespace_after information
                    for i in range(current_multiword_last_token - current_multiword_first_token):
                        # print(i)
                        sentence[-(i + 1)].whitespace_after = 0

                sentence.add_token(token)

            line = file.readline()
        return sentence


class UD_ENGLISH(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_English-EWT/master"
        cached_path(f"{web_path}/en_ewt-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/en_ewt-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/en_ewt-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_ENGLISH, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_GALICIAN(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Galician-TreeGal/master"
        cached_path(f"{web_path}/gl_treegal-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/gl_treegal-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_GALICIAN, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_ANCIENT_GREEK(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Ancient_Greek-PROIEL/master"
        cached_path(f"{web_path}/grc_proiel-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/grc_proiel-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/grc_proiel-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_ANCIENT_GREEK, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_KAZAKH(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Kazakh-KTB/master"
        cached_path(f"{web_path}/kk_ktb-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/kk_ktb-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_KAZAKH, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_OLD_CHURCH_SLAVONIC(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Old_Church_Slavonic-PROIEL/master"
        cached_path(f"{web_path}/cu_proiel-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/cu_proiel-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/cu_proiel-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_OLD_CHURCH_SLAVONIC, self).__init__(
            data_folder, in_memory=in_memory, split_multiwords=split_multiwords
        )


class UD_ARMENIAN(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Armenian-ArmTDP/master/"
        cached_path(f"{web_path}/hy_armtdp-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/hy_armtdp-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/hy_armtdp-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_ARMENIAN, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_ESTONIAN(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Estonian-EDT/master"
        cached_path(f"{web_path}/et_edt-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/et_edt-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/et_edt-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_ESTONIAN, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_GERMAN(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_German-GSD/master"
        cached_path(f"{ud_path}/de_gsd-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/de_gsd-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/de_gsd-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_GERMAN, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_GERMAN_HDT(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = False,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_German-HDT/dev"
        cached_path(f"{ud_path}/de_hdt-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/de_hdt-ud-test.conllu", Path("datasets") / dataset_name)

        train_filenames = [
            "de_hdt-ud-train-a-1.conllu",
            "de_hdt-ud-train-a-2.conllu",
            "de_hdt-ud-train-b-1.conllu",
            "de_hdt-ud-train-b-2.conllu",
        ]

        for train_file in train_filenames:
            cached_path(f"{ud_path}/{train_file}", Path("datasets") / dataset_name / "original")

        data_path = flair.cache_root / "datasets" / dataset_name

        new_train_file: Path = data_path / "de_hdt-ud-train-all.conllu"

        if not new_train_file.is_file():
            with open(new_train_file, "wt") as f_out:
                for train_filename in train_filenames:
                    with open(data_path / "original" / train_filename, "rt") as f_in:
                        f_out.write(f_in.read())

        super(UD_GERMAN_HDT, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_DUTCH(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Dutch-Alpino/master"
        cached_path(f"{ud_path}/nl_alpino-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/nl_alpino-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/nl_alpino-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_DUTCH, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_FAROESE(UniversalDependenciesCorpus):
    """This treebank includes the Faroese treebank dataset from the following link:
    https://github.com/UniversalDependencies/UD_Faroese-FarPaHC/tree/master

    Faronese is a small Western Scandinavian language with 60.000-100.000, related to Icelandic and Old Norse"""

    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Faroese-FarPaHC/master"
        cached_path(f"{web_path}/fo_farpahc-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/fo_farpahc-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/fo_farpahc-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_FAROESE, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_FRENCH(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_French-GSD/master"
        cached_path(f"{ud_path}/fr_gsd-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/fr_gsd-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/fr_gsd-ud-train.conllu", Path("datasets") / dataset_name)
        super(UD_FRENCH, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_ITALIAN(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Italian-ISDT/master"
        cached_path(f"{ud_path}/it_isdt-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/it_isdt-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/it_isdt-ud-train.conllu", Path("datasets") / dataset_name)
        super(UD_ITALIAN, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_LATIN(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Latin-LLCT/master/"
        cached_path(f"{web_path}/la_llct-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/la_llct-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/la_llct-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_LATIN, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_SPANISH(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Spanish-GSD/master"
        cached_path(f"{ud_path}/es_gsd-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/es_gsd-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/es_gsd-ud-train.conllu", Path("datasets") / dataset_name)
        super(UD_SPANISH, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_PORTUGUESE(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Portuguese-Bosque/master"
        cached_path(f"{ud_path}/pt_bosque-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/pt_bosque-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/pt_bosque-ud-train.conllu", Path("datasets") / dataset_name)
        super(UD_PORTUGUESE, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_ROMANIAN(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Romanian-RRT/master"
        cached_path(f"{ud_path}/ro_rrt-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/ro_rrt-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/ro_rrt-ud-train.conllu", Path("datasets") / dataset_name)
        super(UD_ROMANIAN, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_CATALAN(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Catalan-AnCora/master"
        cached_path(f"{ud_path}/ca_ancora-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/ca_ancora-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/ca_ancora-ud-train.conllu", Path("datasets") / dataset_name)
        super(UD_CATALAN, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_POLISH(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Polish-LFG/master"
        cached_path(f"{ud_path}/pl_lfg-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/pl_lfg-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/pl_lfg-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_POLISH, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_CZECH(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = False,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Czech-PDT/master"
        cached_path(f"{ud_path}/cs_pdt-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/cs_pdt-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(
            f"{ud_path}/cs_pdt-ud-train-c.conllu",
            Path("datasets") / dataset_name / "original",
        )
        cached_path(
            f"{ud_path}/cs_pdt-ud-train-l.conllu",
            Path("datasets") / dataset_name / "original",
        )
        cached_path(
            f"{ud_path}/cs_pdt-ud-train-m.conllu",
            Path("datasets") / dataset_name / "original",
        )
        cached_path(
            f"{ud_path}/cs_pdt-ud-train-v.conllu",
            Path("datasets") / dataset_name / "original",
        )
        data_path = flair.cache_root / "datasets" / dataset_name

        train_filenames = [
            "cs_pdt-ud-train-c.conllu",
            "cs_pdt-ud-train-l.conllu",
            "cs_pdt-ud-train-m.conllu",
            "cs_pdt-ud-train-v.conllu",
        ]

        new_train_file: Path = data_path / "cs_pdt-ud-train-all.conllu"

        if not new_train_file.is_file():
            with open(new_train_file, "wt") as f_out:
                for train_filename in train_filenames:
                    with open(data_path / "original" / train_filename, "rt") as f_in:
                        f_out.write(f_in.read())
        super(UD_CZECH, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_SLOVAK(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Slovak-SNK/master"
        cached_path(f"{ud_path}/sk_snk-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/sk_snk-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/sk_snk-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_SLOVAK, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_SWEDISH(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Swedish-Talbanken/master"
        cached_path(f"{ud_path}/sv_talbanken-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/sv_talbanken-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/sv_talbanken-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_SWEDISH, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_DANISH(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Danish-DDT/master"
        cached_path(f"{ud_path}/da_ddt-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/da_ddt-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/da_ddt-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_DANISH, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_NORWEGIAN(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Norwegian-Bokmaal/master"
        cached_path(f"{ud_path}/no_bokmaal-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/no_bokmaal-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/no_bokmaal-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_NORWEGIAN, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_FINNISH(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Finnish-TDT/master"
        cached_path(f"{ud_path}/fi_tdt-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/fi_tdt-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/fi_tdt-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_FINNISH, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_SLOVENIAN(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Slovenian-SSJ/master"
        cached_path(f"{ud_path}/sl_ssj-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/sl_ssj-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/sl_ssj-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_SLOVENIAN, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_CROATIAN(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Croatian-SET/master"
        cached_path(f"{ud_path}/hr_set-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/hr_set-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/hr_set-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_CROATIAN, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_SERBIAN(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Serbian-SET/master"
        cached_path(f"{ud_path}/sr_set-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/sr_set-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/sr_set-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_SERBIAN, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_BULGARIAN(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Bulgarian-BTB/master"
        cached_path(f"{ud_path}/bg_btb-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/bg_btb-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/bg_btb-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_BULGARIAN, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_ARABIC(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Arabic-PADT/master"
        cached_path(f"{ud_path}/ar_padt-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/ar_padt-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/ar_padt-ud-train.conllu", Path("datasets") / dataset_name)
        super(UD_ARABIC, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_HEBREW(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Hebrew-HTB/master"
        cached_path(f"{ud_path}/he_htb-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/he_htb-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/he_htb-ud-train.conllu", Path("datasets") / dataset_name)
        super(UD_HEBREW, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_TURKISH(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        # default dataset folder is the cache root
        if not base_path:
            base_path = flair.cache_root / "datasets"
        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Turkish-IMST/master"
        cached_path(f"{ud_path}/tr_imst-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/tr_imst-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/tr_imst-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_TURKISH, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_PERSIAN(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        # default dataset folder is the cache root
        if not base_path:
            base_path = flair.cache_root / "datasets"
        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Persian-Seraji/master"
        cached_path(f"{ud_path}/fa_seraji-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/fa_seraji-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/fa_seraji-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_PERSIAN, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_RUSSIAN(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Russian-SynTagRus/master"
        cached_path(f"{ud_path}/ru_syntagrus-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/ru_syntagrus-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/ru_syntagrus-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_RUSSIAN, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_HINDI(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Hindi-HDTB/master"
        cached_path(f"{ud_path}/hi_hdtb-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/hi_hdtb-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/hi_hdtb-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_HINDI, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_INDONESIAN(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Indonesian-GSD/master"
        cached_path(f"{ud_path}/id_gsd-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/id_gsd-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/id_gsd-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_INDONESIAN, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_JAPANESE(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Japanese-GSD/master"
        cached_path(f"{ud_path}/ja_gsd-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/ja_gsd-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/ja_gsd-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_JAPANESE, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_CHINESE(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Chinese-GSD/master"
        cached_path(f"{ud_path}/zh_gsd-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/zh_gsd-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/zh_gsd-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_CHINESE, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_KOREAN(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Korean-Kaist/master"
        cached_path(f"{ud_path}/ko_kaist-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/ko_kaist-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/ko_kaist-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_KOREAN, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_BASQUE(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        ud_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Basque-BDT/master"
        cached_path(f"{ud_path}/eu_bdt-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/eu_bdt-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{ud_path}/eu_bdt-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_BASQUE, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_CHINESE_KYOTO(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Classical_Chinese-Kyoto/master"
        cached_path(f"{web_path}/lzh_kyoto-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/lzh_kyoto-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/lzh_kyoto-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_CHINESE_KYOTO, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_GREEK(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Greek-GDT/master"
        cached_path(f"{web_path}/el_gdt-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/el_gdt-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/el_gdt-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_GREEK, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_NAIJA(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Naija-NSC/master"
        cached_path(f"{web_path}//pcm_nsc-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}//pcm_nsc-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}//pcm_nsc-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_NAIJA, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_LIVVI(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Livvi-KKPP/master"
        cached_path(f"{web_path}/olo_kkpp-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/olo_kkpp-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_LIVVI, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_BURYAT(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Buryat-BDT/master"
        cached_path(f"{web_path}/bxr_bdt-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/bxr_bdt-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_BURYAT, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_NORTH_SAMI(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_North_Sami-Giella/master"
        cached_path(f"{web_path}/sme_giella-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/sme_giella-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_NORTH_SAMI, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_MARATHI(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Marathi-UFAL/master"
        cached_path(f"{web_path}/mr_ufal-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/mr_ufal-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/mr_ufal-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_MARATHI, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_MALTESE(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Maltese-MUDT/master"
        cached_path(f"{web_path}/mt_mudt-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/mt_mudt-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/mt_mudt-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_MALTESE, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_AFRIKAANS(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Afrikaans-AfriBooms/master"
        cached_path(f"{web_path}/af_afribooms-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/af_afribooms-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/af_afribooms-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_AFRIKAANS, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_GOTHIC(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Gothic-PROIEL/master"
        cached_path(f"{web_path}/got_proiel-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/got_proiel-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/got_proiel-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_GOTHIC, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_OLD_FRENCH(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Old_French-SRCMF/master"
        cached_path(f"{web_path}/fro_srcmf-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/fro_srcmf-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/fro_srcmf-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_OLD_FRENCH, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_WOLOF(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Wolof-WTB/master"
        cached_path(f"{web_path}/wo_wtb-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/wo_wtb-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/wo_wtb-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_WOLOF, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_BELARUSIAN(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Belarusian-HSE/master"
        cached_path(f"{web_path}/be_hse-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/be_hse-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/be_hse-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_BELARUSIAN, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_COPTIC(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Coptic-Scriptorium/master"
        cached_path(f"{web_path}/cop_scriptorium-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(
            f"{web_path}/cop_scriptorium-ud-test.conllu",
            Path("datasets") / dataset_name,
        )
        cached_path(
            f"{web_path}/cop_scriptorium-ud-train.conllu",
            Path("datasets") / dataset_name,
        )

        super(UD_COPTIC, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_IRISH(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Irish-IDT/master"
        cached_path(f"{web_path}/ga_idt-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/ga_idt-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/ga_idt-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_IRISH, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_LATVIAN(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):

        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Latvian-LVTB/master"
        cached_path(f"{web_path}/lv_lvtb-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/lv_lvtb-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/lv_lvtb-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_LATVIAN, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)


class UD_LITHUANIAN(UniversalDependenciesCorpus):
    def __init__(
        self,
        base_path: Union[str, Path] = None,
        in_memory: bool = True,
        split_multiwords: bool = True,
    ):
        if not base_path:
            base_path = Path(flair.cache_root) / "datasets"
        else:
            base_path = Path(base_path)

        # this dataset name
        dataset_name = self.__class__.__name__.lower()

        # default dataset folder is the cache root

        data_folder = base_path / dataset_name

        # download data if necessary
        web_path = "https://raw.githubusercontent.com/UniversalDependencies/UD_Lithuanian-ALKSNIS/master"
        cached_path(f"{web_path}/lt_alksnis-ud-dev.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/lt_alksnis-ud-test.conllu", Path("datasets") / dataset_name)
        cached_path(f"{web_path}/lt_alksnis-ud-train.conllu", Path("datasets") / dataset_name)

        super(UD_LITHUANIAN, self).__init__(data_folder, in_memory=in_memory, split_multiwords=split_multiwords)



from collections import defaultdict
from typing import Dict


def get_spans_from_bio(bioes_tags, bioes_scores=None):
    # add a dummy "O" to close final prediction
    bioes_tags.append("O")
    # return complex list
    found_spans = []
    # internal variables
    current_tag_weights: Dict[str, float] = defaultdict(lambda: 0.0)
    previous_tag = "O-"
    current_span = []
    current_span_scores = []
    for idx, bioes_tag in enumerate(bioes_tags):

        # non-set tags are OUT tags
        if bioes_tag == "" or bioes_tag == "O" or bioes_tag == "_":
            bioes_tag = "O-"

        # anything that is not OUT is IN
        in_span = False if bioes_tag == "O-" else True

        # does this prediction start a new span?
        starts_new_span = False

        # begin and single tags start new spans
        if bioes_tag[0:2] in ["B-", "S-"]:
            starts_new_span = True

        # in IOB format, an I tag starts a span if it follows an O or is a different span
        if bioes_tag[0:2] == "I-" and previous_tag[2:] != bioes_tag[2:]:
            starts_new_span = True

        # single tags that change prediction start new spans
        if bioes_tag[0:2] in ["S-"] and previous_tag[2:] != bioes_tag[2:]:
            starts_new_span = True

        # if an existing span is ended (either by reaching O or starting a new span)
        if (starts_new_span or not in_span) and len(current_span) > 0:
            # determine score and value
            span_score = sum(current_span_scores) / len(current_span_scores)
            span_value = sorted(current_tag_weights.items(), key=lambda k_v: k_v[1], reverse=True)[0][0]

            # append to result list
            found_spans.append((current_span, span_score, span_value))

            # reset for-loop variables for new span
            current_span = []
            current_span_scores = []
            current_tag_weights = defaultdict(lambda: 0.0)

        if in_span:
            current_span.append(idx)
            current_span_scores.append(bioes_scores[idx] if bioes_scores else 1.0)
            weight = 1.1 if starts_new_span else 1.0
            current_tag_weights[bioes_tag[2:]] += weight

        # remember previous tag
        previous_tag = bioes_tag

    return found_spans



import copy
import shutil

import pytest

import flair
import flair.datasets
from flair.data import MultiCorpus, Sentence
from flair.datasets import ColumnCorpus
from flair.datasets.sequence_labeling import (
    JsonlCorpus,
    JsonlDataset,
    MultiFileJsonlCorpus,
)


def test_load_imdb_data(tasks_base_path):
    # get training, test and dev data
    corpus = flair.datasets.ClassificationCorpus(
        tasks_base_path / "imdb",
        memory_mode="full",
    )

    assert len(corpus.train) == 5
    assert len(corpus.dev) == 5
    assert len(corpus.test) == 5


def test_load_imdb_data_streaming(tasks_base_path):
    # get training, test and dev data
    corpus = flair.datasets.ClassificationCorpus(
        tasks_base_path / "imdb",
        memory_mode="disk",
    )

    assert len(corpus.train) == 5
    assert len(corpus.dev) == 5
    assert len(corpus.test) == 5


def test_load_imdb_data_max_tokens(tasks_base_path):
    # get training, test and dev data
    corpus = flair.datasets.ClassificationCorpus(tasks_base_path / "imdb", memory_mode="full", truncate_to_max_tokens=3)

    assert len(corpus.train[0]) <= 3
    assert len(corpus.dev[0]) <= 3
    assert len(corpus.test[0]) <= 3


def test_load_imdb_data_streaming_max_tokens(tasks_base_path):
    # get training, test and dev data
    corpus = flair.datasets.ClassificationCorpus(tasks_base_path / "imdb", memory_mode="full", truncate_to_max_tokens=3)

    assert len(corpus.train[0]) <= 3
    assert len(corpus.dev[0]) <= 3
    assert len(corpus.test[0]) <= 3


def test_load_ag_news_data(tasks_base_path):
    # get training, test and dev data
    corpus = flair.datasets.ClassificationCorpus(tasks_base_path / "ag_news")

    assert len(corpus.train) == 10
    assert len(corpus.dev) == 10
    assert len(corpus.test) == 10


def test_load_sequence_labeling_data(tasks_base_path):
    # get training, test and dev data
    corpus = flair.datasets.ColumnCorpus(tasks_base_path / "fashion", column_format={0: "text", 2: "ner"})

    assert len(corpus.train) == 6
    assert len(corpus.dev) == 1
    assert len(corpus.test) == 1


def test_load_sequence_labeling_whitespace_after(tasks_base_path):
    # get training, test and dev data
    corpus = flair.datasets.ColumnCorpus(
        tasks_base_path / "column_with_whitespaces",
        column_format={0: "text", 1: "ner", 2: "space-after"},
    )

    assert len(corpus.train) == 1
    assert len(corpus.dev) == 1
    assert len(corpus.test) == 1

    assert corpus.train[0].to_tokenized_string() == "It is a German - owned firm ."
    assert corpus.train[0].to_plain_string() == "It is a German-owned firm."
    for token in corpus.train[0]:
        assert token.start_pos is not None
        assert token.end_pos is not None


def test_load_column_corpus_options(tasks_base_path):
    # get training, test and dev data
    corpus = flair.datasets.ColumnCorpus(
        tasks_base_path / "column_corpus_options",
        column_format={0: "text", 1: "ner"},
        column_delimiter="\t",
        skip_first_line=True,
    )

    assert len(corpus.train) == 1
    assert len(corpus.dev) == 1
    assert len(corpus.test) == 1

    assert corpus.train[0].to_tokenized_string() == "This is New Berlin"


def test_load_span_data(tasks_base_path):
    # load column dataset with one entry
    dataset = flair.datasets.ColumnDataset(
        tasks_base_path / "span_labels" / "span_first.txt",
        column_name_map={0: "text", 1: "ner"},
    )

    assert len(dataset) == 1
    assert dataset[0][2].text == "RAB"
    assert dataset[0][2].get_label("ner").value == "PARTA"

    # load column dataset with two entries
    dataset = flair.datasets.ColumnDataset(
        tasks_base_path / "span_labels" / "span_second.txt",
        column_name_map={0: "text", 1: "ner"},
    )

    assert len(dataset) == 2
    assert dataset[1][2].text == "RAB"
    assert dataset[1][2].get_label("ner").value == "PARTA"

    # load column dataset with three entries
    dataset = flair.datasets.ColumnDataset(
        tasks_base_path / "span_labels" / "span_third.txt",
        column_name_map={0: "text", 1: "ner"},
    )

    assert len(dataset) == 3
    assert dataset[2][2].text == "RAB"
    assert dataset[2][2].get_label("ner").value == "PARTA"


def test_load_germeval_data(tasks_base_path):
    # get training, test and dev data
    corpus = flair.datasets.ColumnCorpus(tasks_base_path / "ner_german_germeval", column_format={0: "text", 2: "ner"})

    assert len(corpus.train) == 2
    assert len(corpus.dev) == 1
    assert len(corpus.test) == 1


def test_load_ud_english_data(tasks_base_path):
    # get training, test and dev data
    corpus = flair.datasets.UD_ENGLISH(tasks_base_path)

    assert len(corpus.train) == 6
    assert len(corpus.test) == 4
    assert len(corpus.dev) == 2

    # check if Token labels are correct
    sentence = corpus.train[0]
    assert sentence[0].text == "From"
    assert sentence[0].get_label("upos").value == "ADP"
    assert sentence[1].text == "the"
    assert sentence[1].get_label("upos").value == "DET"


def test_load_up_english_data(tasks_base_path):
    # get training, test and dev data
    corpus = flair.datasets.UP_ENGLISH(tasks_base_path)

    assert len(corpus.train) == 4
    assert len(corpus.test) == 2
    assert len(corpus.dev) == 2

    # check if Token labels for frames are correct
    sentence = corpus.dev[0]
    assert sentence[2].text == "AP"
    assert sentence[2].get_label("frame", zero_tag_value="no_label").value == "no_label"
    assert sentence[3].text == "comes"
    assert sentence[3].get_label("frame").value == "come.03"


def test_load_no_dev_data(tasks_base_path):
    # get training, test and dev data
    corpus = flair.datasets.ColumnCorpus(tasks_base_path / "fashion_nodev", column_format={0: "text", 2: "ner"})

    assert len(corpus.train) == 5
    assert len(corpus.dev) == 1
    assert len(corpus.test) == 1


def test_load_no_dev_data_explicit(tasks_base_path):
    # get training, test and dev data
    corpus = flair.datasets.ColumnCorpus(
        tasks_base_path / "fashion_nodev",
        column_format={0: "text", 2: "ner"},
        train_file="train.tsv",
        test_file="test.tsv",
    )

    assert len(corpus.train) == 5
    assert len(corpus.dev) == 1
    assert len(corpus.test) == 1


def test_multi_corpus(tasks_base_path):
    corpus_1 = flair.datasets.ColumnCorpus(tasks_base_path / "ner_german_germeval", column_format={0: "text", 2: "ner"})

    corpus_2 = flair.datasets.ColumnCorpus(tasks_base_path / "fashion", column_format={0: "text", 2: "ner"})
    # get two corpora as one
    corpus = MultiCorpus([corpus_1, corpus_2])

    assert len(corpus.train) == 8
    assert len(corpus.dev) == 2
    assert len(corpus.test) == 2


def test_download_load_data(tasks_base_path):
    # get training, test and dev data for full English UD corpus from web
    corpus = flair.datasets.UD_ENGLISH()

    assert len(corpus.train) == 12543
    assert len(corpus.dev) == 2001
    assert len(corpus.test) == 2077

    # clean up data directory
    shutil.rmtree(flair.cache_root / "datasets" / "ud_english")


def _assert_conllu_dataset(dataset):
    sent1 = dataset[0]

    assert [label.data_point.text for label in sent1.get_labels("ner")] == ["Larry Page", "Sergey Brin", "Google"]
    assert [label.value for label in sent1.get_labels("ner")] == ["PER", "PER", "ORG"]

    assert [token.get_label("upos").value for token in sent1.tokens] == [
        "PROPN",
        "PROPN",
        "CCONJ",
        "PROPN",
        "PROPN",
        "VERB",
        "PROPN",
        "PUNCT",
    ]

    assert [token.whitespace_after for token in sent1.tokens] == [
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        1,
    ]

    ner_spans1 = sent1.get_labels("ner")
    assert len(ner_spans1) == 3

    upos_spans1 = sent1.get_labels("upos")
    assert len(upos_spans1) == 8

    rels1 = sent1.get_labels("relation")
    assert len(rels1) == 2

    assert [token.idx for token in rels1[1].data_point.first] == [7]
    assert [token.idx for token in rels1[1].data_point.second] == [4, 5]

    sent3 = dataset[2]

    ner_labels3 = sent3.get_labels("ner")
    assert len(ner_labels3) == 3

    upos_labels3 = sent3.get_labels("upos")
    assert len(upos_labels3) == 11

    rels3 = sent3.get_labels("relation")
    assert len(rels3) == 1

    assert [token.idx for token in rels3[0].data_point.first] == [6]
    assert [token.idx for token in rels3[0].data_point.second] == [1, 2]


def test_load_conllu_corpus(tasks_base_path):
    corpus = ColumnCorpus(
        tasks_base_path / "conllu",
        train_file="train.conllu",
        dev_file="train.conllu",
        test_file="train.conllu",
        in_memory=False,
        column_format={1: "text", 2: "upos", 3: "ner", 4: "feats"},
    )

    assert len(corpus.train) == 4
    assert len(corpus.dev) == 4
    assert len(corpus.test) == 4

    _assert_conllu_dataset(corpus.train)


def test_load_conllu_corpus_in_memory(tasks_base_path):
    corpus = ColumnCorpus(
        tasks_base_path / "conllu",
        train_file="train.conllu",
        dev_file="train.conllu",
        test_file="train.conllu",
        column_format={1: "text", 2: "upos", 3: "ner", 4: "feats"},
        in_memory=True,
    )

    assert len(corpus.train) == 4
    assert len(corpus.dev) == 4
    assert len(corpus.test) == 4

    _assert_conllu_dataset(corpus.train)


def test_load_conllu_plus_corpus(tasks_base_path):
    corpus = ColumnCorpus(
        tasks_base_path / "conllu",
        train_file="train.conllup",
        dev_file="train.conllup",
        test_file="train.conllup",
        column_format={1: "text", 2: "upos", 3: "ner", 4: "feats"},
        in_memory=False,
    )

    assert len(corpus.train) == 4
    assert len(corpus.dev) == 4
    assert len(corpus.test) == 4

    _assert_conllu_dataset(corpus.train)


def test_load_conllu_corpus_plus_in_memory(tasks_base_path):
    corpus = ColumnCorpus(
        tasks_base_path / "conllu",
        train_file="train.conllup",
        dev_file="train.conllup",
        test_file="train.conllup",
        column_format={1: "text", 2: "upos", 3: "ner", 4: "feats"},
        in_memory=True,
    )

    assert len(corpus.train) == 4
    assert len(corpus.dev) == 4
    assert len(corpus.test) == 4

    _assert_conllu_dataset(corpus.train)


def _assert_universal_dependencies_conllu_dataset(dataset):
    sent1: Sentence = dataset[0]

    assert [token.whitespace_after for token in sent1.tokens] == [
        1,
        1,
        1,
        1,
        0,
        1,
    ]

    assert len(sent1.get_labels("Number")) == 4
    assert sent1[1].get_labels("Number")[0].value == "Plur"
    assert sent1[1].get_labels("Person")[0].value == "3"
    assert sent1[1].get_labels("Tense")[0].value == "Pres"

    # assert [token.get_tag("head").value for token in sent1.tokens] == [
    #     "2",
    #     "0",
    #     "4",
    #     "2",
    #     "2",
    #     "2",
    # ]

    assert [token.get_label("deprel").value for token in sent1.tokens] == [
        "nsubj",
        "root",
        "cc",
        "conj",
        "obj",
        "punct",
    ]


def test_load_universal_dependencies_conllu_corpus(tasks_base_path):
    """
    This test only covers basic universal dependencies datasets.
    For example, multi-word tokens or the "deps" column sentence annotations
    are not supported yet.
    """

    # Here, we use the default token annotation fields.
    corpus = ColumnCorpus(
        tasks_base_path / "conllu",
        train_file="universal_dependencies.conllu",
        dev_file="universal_dependencies.conllu",
        test_file="universal_dependencies.conllu",
        column_format={
            1: "text",
            2: "lemma",
            3: "upos",
            4: "pos",
            5: "feats",
            6: "head",
            7: "deprel",
            8: "deps",
            9: "misc",
        },
    )

    assert len(corpus.train) == 1
    assert len(corpus.dev) == 1
    assert len(corpus.test) == 1

    _assert_universal_dependencies_conllu_dataset(corpus.train)


def test_hipe_2022_corpus(tasks_base_path):
    """
    This test covers the complete v1.0 version of the HIPE 2022,
    including the version with document separator.
    """

    # We have manually checked, that these numbers are correct:
    hipe_stats = {
        "v1.0": {
            "ajmc": {
                "de": {"sample": {"sents": 119, "docs": 8, "labels": ["date", "loc", "pers", "scope", "work"]}},
                "en": {"sample": {"sents": 83, "docs": 5, "labels": ["date", "loc", "pers", "scope", "work"]}},
            },
            "hipe2020": {
                "de": {
                    "train": {
                        "sents": 3470 + 2,  # 2 sentences with missing EOS marker
                        "docs": 103,
                        "labels": ["loc", "org", "pers", "prod", "time"],
                    },
                    "dev": {"sents": 1202, "docs": 33, "labels": ["loc", "org", "pers", "prod", "time"]},
                },
                "en": {"dev": {"sents": 1045, "docs": 80, "labels": ["loc", "org", "pers", "prod", "time"]}},
                "fr": {
                    "train": {"sents": 5743, "docs": 158, "labels": ["loc", "org", "pers", "prod", "time", "comp"]},
                    "dev": {"sents": 1244, "docs": 43, "labels": ["loc", "org", "pers", "prod", "time"]},
                },
            },
            "letemps": {
                "fr": {
                    "train": {"sents": 14051, "docs": 414, "labels": ["loc", "org", "pers"]},
                    "dev": {"sents": 1341, "docs": 51, "labels": ["loc", "org", "pers"]},
                }
            },
            "newseye": {
                # +1 offset, because of missing EOS marker at EOD
                "de": {
                    "train": {"sents": 23646 + 1, "docs": 11, "labels": ["HumanProd", "LOC", "ORG", "PER"]},
                    "dev": {"sents": 1110 + 1, "docs": 12, "labels": ["HumanProd", "LOC", "ORG", "PER"]},
                    "dev2": {"sents": 1541 + 1, "docs": 12, "labels": ["HumanProd", "LOC", "ORG", "PER"]},
                },
                "fi": {
                    "train": {"sents": 1141 + 1, "docs": 24, "labels": ["HumanProd", "LOC", "ORG", "PER"]},
                    "dev": {"sents": 140 + 1, "docs": 24, "labels": ["HumanProd", "LOC", "ORG", "PER"]},
                    "dev2": {"sents": 104 + 1, "docs": 21, "labels": ["HumanProd", "LOC", "ORG", "PER"]},
                },
                "fr": {
                    "train": {"sents": 7106 + 1, "docs": 35, "labels": ["HumanProd", "LOC", "ORG", "PER"]},
                    "dev": {"sents": 662 + 1, "docs": 35, "labels": ["HumanProd", "LOC", "ORG", "PER"]},
                    "dev2": {"sents": 1016 + 1, "docs": 35, "labels": ["HumanProd", "LOC", "ORG", "PER"]},
                },
                "sv": {
                    "train": {"sents": 1063 + 1, "docs": 21, "labels": ["HumanProd", "LOC", "ORG", "PER"]},
                    "dev": {"sents": 126 + 1, "docs": 21, "labels": ["HumanProd", "LOC", "ORG", "PER"]},
                    "dev2": {"sents": 136 + 1, "docs": 21, "labels": ["HumanProd", "LOC", "ORG", "PER"]},
                },
            },
            "sonar": {
                "de": {
                    "dev": {
                        "sents": 1603 + 10,  # 10 sentences with missing EOS marker
                        "docs": 10,
                        "labels": ["LOC", "ORG", "PER"],
                    }
                }
            },
            "topres19th": {
                "en": {
                    "train": {"sents": 5874, "docs": 309, "labels": ["BUILDING", "LOC", "STREET"]},
                    "dev": {"sents": 646, "docs": 34, "labels": ["BUILDING", "LOC", "STREET"]},
                }
            },
        }
    }

    hipe_stats["v2.0"] = hipe_stats["v1.0"].copy()
    hipe_stats["v2.0"]["ajmc"] = {
        "de": {
            "train": {
                "sents": 1022 + 2,  # 2 sentences with missing EOS marker
                "docs": 76,
                "labels": ["date", "loc", "object", "pers", "scope", "work"],
            },
            "dev": {"sents": 192, "docs": 14, "labels": ["loc", "object", "pers", "scope", "work"]},
        },
        "en": {
            "train": {
                "sents": 1153 + 1,  # 1 sentence with missing EOS marker
                "docs": 60,
                "labels": ["date", "loc", "object", "pers", "scope", "work"],
            },
            "dev": {
                "sents": 251 + 1,  # 1 sentence with missing EOS marker
                "docs": 14,
                "labels": ["date", "loc", "pers", "scope", "work"],
            },
        },
        "fr": {
            "train": {
                "sents": 893 + 1,  # 1 sentence with missing EOS marker
                "docs": 72,
                "labels": ["date", "loc", "object", "pers", "scope", "work"],
            },
            "dev": {
                "sents": 201 + 1,  # 1 sentence with missing EOS marker
                "docs": 17,
                "labels": ["pers", "scope", "work"],
            },
        },
    }
    hipe_stats["v2.0"]["newseye"] = {
        "de": {
            "train": {"sents": 20839 + 1, "docs": 7, "labels": ["HumanProd", "LOC", "ORG", "PER"]}  # missing EOD marker
        }
    }
    hipe_stats["v2.0"]["sonar"] = {
        "de": {
            "dev": {
                "sents": 816 + 10,  # 9 sentences with missing EOS marker + missing EOD
                "docs": 10,
                "labels": ["LOC", "ORG", "PER"],
            }
        }
    }
    hipe_stats["v2.1"] = copy.deepcopy(hipe_stats["v2.0"])
    hipe_stats["v2.1"]["hipe2020"]["fr"]["train"] = {
        "sents": 5743,
        "docs": 158,
        "labels": ["loc", "org", "pers", "prod", "time"],
    }

    def test_hipe_2022(dataset_version="v2.1", add_document_separator=True):
        for dataset_name, languages in hipe_stats[dataset_version].items():
            for language in languages:
                splits = languages[language]

                corpus = flair.datasets.NER_HIPE_2022(
                    version=dataset_version,
                    dataset_name=dataset_name,
                    language=language,
                    dev_split_name="dev",
                    add_document_separator=add_document_separator,
                )

                for split_name, stats in splits.items():
                    split_description = f"{dataset_name}@{dataset_version}/{language}#{split_name}"

                    current_sents = stats["sents"]
                    current_docs = stats["docs"]
                    current_labels = set(stats["labels"] + ["<unk>"])

                    total_sentences = current_sents + current_docs if add_document_separator else stats["sents"]

                    if split_name == "train":
                        assert (
                            len(corpus.train) == total_sentences
                        ), f"Sentence count mismatch for {split_description}: {len(corpus.train)} vs. {total_sentences}"

                        gold_labels = set(corpus.make_label_dictionary(label_type="ner").get_items())

                        assert (
                            current_labels == gold_labels
                        ), f"Label mismatch for {split_description}: {current_labels} vs. {gold_labels}"

                    elif split_name in ["dev", "sample"]:
                        assert (
                            len(corpus.dev) == total_sentences
                        ), f"Sentence count mismatch for {split_description}: {len(corpus.dev)} vs. {total_sentences}"

                        corpus._train = corpus._dev
                        gold_labels = set(corpus.make_label_dictionary(label_type="ner").get_items())

                        assert (
                            current_labels == gold_labels
                        ), f"Label mismatch for {split_description}: {current_labels} vs. {gold_labels}"
                    elif split_name == "dev2":
                        corpus = flair.datasets.NER_HIPE_2022(
                            version=dataset_version,
                            dataset_name=dataset_name,
                            language=language,
                            dev_split_name="dev2",
                            add_document_separator=add_document_separator,
                        )

                        corpus._train = corpus._dev
                        gold_labels = set(corpus.make_label_dictionary(label_type="ner").get_items())

                        assert (
                            len(corpus.dev) == total_sentences
                        ), f"Sentence count mismatch for {split_description}: {len(corpus.dev)} vs. {total_sentences}"

                        assert (
                            current_labels == gold_labels
                        ), f"Label mismatch for {split_description}: {current_labels} vs. {gold_labels}"

    test_hipe_2022(dataset_version="v1.0", add_document_separator=True)
    test_hipe_2022(dataset_version="v1.0", add_document_separator=False)
    test_hipe_2022(dataset_version="v2.0", add_document_separator=True)
    test_hipe_2022(dataset_version="v2.0", add_document_separator=False)
    test_hipe_2022(dataset_version="v2.1", add_document_separator=True)
    test_hipe_2022(dataset_version="v2.1", add_document_separator=False)


def test_multi_file_jsonl_corpus_should_use_label_type(tasks_base_path):
    corpus = MultiFileJsonlCorpus(
        train_files=[tasks_base_path / "jsonl/train.jsonl"],
        dev_files=[tasks_base_path / "jsonl/testa.jsonl"],
        test_files=[tasks_base_path / "jsonl/testb.jsonl"],
        label_type="pos",
    )

    for sentence in corpus.get_all_sentences():
        assert sentence.has_label("pos")
        assert not sentence.has_label("ner")


def test_jsonl_corpus_should_use_label_type(tasks_base_path):
    corpus = JsonlCorpus(tasks_base_path / "jsonl", label_type="pos")

    for sentence in corpus.get_all_sentences():
        assert sentence.has_label("pos")
        assert not sentence.has_label("ner")


def test_jsonl_dataset_should_use_label_type(tasks_base_path):
    """
    Tests whether the dataset respects the label_type parameter
    """
    dataset = JsonlDataset(tasks_base_path / "jsonl/train.jsonl", label_type="pos")  # use other type

    for sentence in dataset.sentences:
        assert sentence.has_label("pos")
        assert not sentence.has_label("ner")


def test_reading_jsonl_dataset_should_be_successful(tasks_base_path):
    """
    Tests reading a JsonlDataset containing multiple tagged entries
    """
    dataset = JsonlDataset(tasks_base_path / "jsonl/train.jsonl")

    assert len(dataset.sentences) == 5
    assert dataset.sentences[0].get_token(3).get_label("ner").value == "B-LOC"
    assert dataset.sentences[0].get_token(4).get_label("ner").value == "I-LOC"


def test_simple_folder_jsonl_corpus_should_load(tasks_base_path):
    corpus = JsonlCorpus(tasks_base_path / "jsonl")
    assert len(corpus.get_all_sentences()) == 11


TRAIN_FILE = "tests/resources/tasks/jsonl/train.jsonl"
TESTA_FILE = "tests/resources/tasks/jsonl/testa.jsonl"
TESTB_FILE = "tests/resources/tasks/jsonl/testa.jsonl"


@pytest.mark.parametrize(
    "train_files,dev_files,test_files,expected_size",
    [
        (
            [TRAIN_FILE],
            [TESTA_FILE],
            [TESTB_FILE],
            11,
        ),
        (
            [TRAIN_FILE],
            [],
            [TESTB_FILE],
            8,
        ),
        (
            [TRAIN_FILE],
            [],
            None,
            5,
        ),
        (
            None,
            [TESTA_FILE],
            None,
            3,
        ),
        (
            [TRAIN_FILE, TESTA_FILE],
            [TESTA_FILE],
            [TESTB_FILE],
            14,
        ),
    ],
)
def test_corpus_with_single_files_should_load(train_files, dev_files, test_files, expected_size):
    corpus = MultiFileJsonlCorpus(train_files, dev_files, test_files)
    assert len(corpus.get_all_sentences()) == expected_size


def test_empty_corpus_should_raise_error():
    with pytest.raises(RuntimeError) as err:
        MultiFileJsonlCorpus(None, None, None)

    assert str(err.value) == "No data provided when initializing corpus object."



import time
from abc import abstractmethod
from collections import Counter
from typing import List, Union

import torch
from torch.nn.parameter import Parameter

import flair
from flair.data import Sentence
from flair.embeddings import (
    DocumentEmbeddings,
    FlairEmbeddings,
    StackedEmbeddings,
    TokenEmbeddings,
)
from flair.nn import LockedDropout, WordDropout


class TransformerWordEmbeddings(TokenEmbeddings):
    def __init__(
        self,
        model: str = "bert-base-uncased",
        layers: str = "-1,-2,-3,-4",
        pooling_operation: str = "first",
        batch_size: int = 1,
        use_scalar_mix: bool = False,
        fine_tune: bool = False,
    ):
        """
        Bidirectional transformer embeddings of words from various transformer architectures.
        :param model: name of transformer model (see https://huggingface.co/transformers/pretrained_models.html for
        options)
        :param layers: string indicating which layers to take for embedding (-1 is topmost layer)
        :param pooling_operation: how to get from token piece embeddings to token embedding. Either take the first
        subtoken ('first'), the last subtoken ('last'), both first and last ('first_last') or a mean over all ('mean')
        :param batch_size: How many sentence to push through transformer at once. Set to 1 by default since transformer
        models tend to be huge.
        :param use_scalar_mix: If True, uses a scalar mix of layers as embedding
        :param fine_tune: If True, allows transformers to be fine-tuned during training
        """
        super().__init__()

        # load tokenizer and transformer model
        self.tokenizer = AutoTokenizer.from_pretrained(model)
        config = AutoConfig.from_pretrained(model, output_hidden_states=True)
        self.model = AutoModel.from_pretrained(model, config=config)

        # model name
        self.name = "transformer-word-" + str(model)

        # when initializing, embeddings are in eval mode by default
        self.model.eval()
        self.model.to(flair.device)

        # embedding parameters
        if layers == "all":
            # send mini-token through to check how many layers the model has
            hidden_states = self.model(torch.tensor([1], device=flair.device).unsqueeze(0))[-1]
            self.layer_indexes = [int(x) for x in range(len(hidden_states))]
        else:
            self.layer_indexes = [int(x) for x in layers.split(",")]
        self.mix = ScalarMix(mixture_size=len(self.layer_indexes), trainable=False)
        self.pooling_operation = pooling_operation
        self.use_scalar_mix = use_scalar_mix
        self.fine_tune = fine_tune
        self.static_embeddings = not self.fine_tune
        self.batch_size = batch_size

        self.special_tokens = []
        self.special_tokens.append(self.tokenizer.bos_token)
        self.special_tokens.append(self.tokenizer.cls_token)

        # most models have an intial BOS token, except for XLNet, T5 and GPT2
        self.begin_offset = 1
        if type(self.tokenizer) == XLNetTokenizer:
            self.begin_offset = 0
        if type(self.tokenizer) == T5Tokenizer:
            self.begin_offset = 0
        if type(self.tokenizer) == GPT2Tokenizer:
            self.begin_offset = 0

    def _add_embeddings_internal(self, sentences: List[Sentence]) -> List[Sentence]:
        """Add embeddings to all words in a list of sentences."""

        # split into micro batches of size self.batch_size before pushing through transformer
        sentence_batches = [
            sentences[i * self.batch_size : (i + 1) * self.batch_size]
            for i in range((len(sentences) + self.batch_size - 1) // self.batch_size)
        ]

        # embed each micro-batch
        for batch in sentence_batches:
            self._add_embeddings_to_sentences(batch)

        return sentences

    def _add_embeddings_to_sentences(self, sentences: List[Sentence]):
        """Match subtokenization to Flair tokenization and extract embeddings from transformers for each token."""

        # first, subtokenize each sentence and find out into how many subtokens each token was divided
        subtokenized_sentences = []
        subtokenized_sentences_token_lengths = []

        for sentence in sentences:

            tokenized_string = sentence.to_tokenized_string()

            # method 1: subtokenize sentence
            # subtokenized_sentence = self.tokenizer.encode(tokenized_string, add_special_tokens=True)

            # method 2:
            ids = self.tokenizer.encode(tokenized_string, add_special_tokens=False)
            subtokenized_sentence = self.tokenizer.build_inputs_with_special_tokens(ids)

            subtokenized_sentences.append(torch.tensor(subtokenized_sentence, dtype=torch.long))
            subtokens = self.tokenizer.convert_ids_to_tokens(subtokenized_sentence)

            word_iterator = iter(sentence)
            token = next(word_iterator)
            token_text = token.text.lower()

            token_subtoken_lengths = []
            reconstructed_token = ""
            subtoken_count = 0

            # iterate over subtokens and reconstruct tokens
            for subtoken_id, subtoken in enumerate(subtokens):

                subtoken_count += 1

                # remove special markup
                subtoken = re.sub("^Ġ", "", subtoken)  # RoBERTa models
                subtoken = re.sub("^##", "", subtoken)  # BERT models
                subtoken = re.sub("^▁", "", subtoken)  # XLNet models
                subtoken = re.sub("</w>$", "", subtoken)  # XLM models

                # append subtoken to reconstruct token
                reconstructed_token = reconstructed_token + subtoken

                # check if reconstructed token is special begin token ([CLS] or similar)
                if reconstructed_token in self.special_tokens and subtoken_id == 0:
                    reconstructed_token = ""
                    subtoken_count = 0

                # special handling for UNK subtokens
                if self.tokenizer.unk_token and self.tokenizer.unk_token in reconstructed_token:
                    pieces = self.tokenizer.convert_ids_to_tokens(
                        self.tokenizer.encode(token.text, add_special_tokens=False)
                    )
                    token_text = ""
                    for piece in pieces:
                        # remove special markup
                        piece = re.sub("^Ġ", "", piece)  # RoBERTa models
                        piece = re.sub("^##", "", piece)  # BERT models
                        piece = re.sub("^▁", "", piece)  # XLNet models
                        piece = re.sub("</w>$", "", piece)  # XLM models
                        token_text += piece
                    token_text = token_text.lower()

                # check if reconstructed token is the same as current token
                if reconstructed_token.lower() == token_text:

                    # if so, add subtoken count
                    token_subtoken_lengths.append(subtoken_count)

                    # reset subtoken count and reconstructed token
                    reconstructed_token = ""
                    subtoken_count = 0

                    # break from loop if all tokens are accounted for
                    if len(token_subtoken_lengths) < len(sentence):
                        token = next(word_iterator)
                        token_text = token.text.lower()
                    else:
                        break

            subtokenized_sentences_token_lengths.append(token_subtoken_lengths)

        # find longest sentence in batch
        longest_sequence_in_batch: int = len(max(subtokenized_sentences, key=len))

        # initialize batch tensors and mask
        input_ids = torch.zeros(
            [len(sentences), longest_sequence_in_batch],
            dtype=torch.long,
            device=flair.device,
        )
        mask = torch.zeros(
            [len(sentences), longest_sequence_in_batch],
            dtype=torch.long,
            device=flair.device,
        )
        for s_id, sentence in enumerate(subtokenized_sentences):
            sequence_length = len(sentence)
            input_ids[s_id][:sequence_length] = sentence
            mask[s_id][:sequence_length] = torch.ones(sequence_length)

        # put encoded batch through transformer model to get all hidden states of all encoder layers
        hidden_states = self.model(input_ids, attention_mask=mask)[-1]

        # gradients are enabled if fine-tuning is enabled
        gradient_context = torch.enable_grad() if (self.fine_tune and self.training) else torch.no_grad()

        with gradient_context:

            # iterate over all subtokenized sentences
            for sentence_idx, (sentence, subtoken_lengths) in enumerate(
                zip(sentences, subtokenized_sentences_token_lengths)
            ):

                subword_start_idx = self.begin_offset

                # for each token, get embedding
                for token_idx, (token, number_of_subtokens) in enumerate(zip(sentence, subtoken_lengths)):

                    subword_end_idx = subword_start_idx + number_of_subtokens

                    subtoken_embeddings: List[torch.FloatTensor] = []

                    # get states from all selected layers, aggregate with pooling operation
                    for layer in self.layer_indexes:
                        current_embeddings = hidden_states[layer][sentence_idx][subword_start_idx:subword_end_idx]

                        if self.pooling_operation == "first":
                            final_embedding: torch.FloatTensor = current_embeddings[0]

                        if self.pooling_operation == "last":
                            final_embedding: torch.FloatTensor = current_embeddings[-1]

                        if self.pooling_operation == "first_last":
                            final_embedding: torch.Tensor = torch.cat([current_embeddings[0], current_embeddings[-1]])

                        if self.pooling_operation == "mean":
                            all_embeddings: List[torch.FloatTensor] = [
                                embedding.unsqueeze(0) for embedding in current_embeddings
                            ]
                            final_embedding: torch.Tensor = torch.mean(torch.cat(all_embeddings, dim=0), dim=0)

                        subtoken_embeddings.append(final_embedding)

                    # use scalar mix of embeddings if so selected
                    if self.use_scalar_mix:
                        sm_embeddings = torch.mean(torch.stack(subtoken_embeddings, dim=1), dim=1)
                        # sm_embeddings = self.mix(subtoken_embeddings)

                        subtoken_embeddings = [sm_embeddings]

                    # set the extracted embedding for the token
                    token.set_embedding(self.name, torch.cat(subtoken_embeddings))

                    subword_start_idx += number_of_subtokens

    def train(self, mode=True):
        # if fine-tuning is not enabled (i.e. a "feature-based approach" used), this
        # module should never be in training mode
        if not self.fine_tune:
            pass
        else:
            super().train(mode)

    @property
    @abstractmethod
    def embedding_length(self) -> int:
        """Returns the length of the embedding vector."""

        if not self.use_scalar_mix:
            length = len(self.layer_indexes) * self.model.config.hidden_size
        else:
            length = self.model.config.hidden_size

        if self.pooling_operation == "first_last":
            length *= 2

        return length


class Stopwatch(object):
    """A stopwatch utility for timing execution that can be used as a regular
    object or as a context manager.
    NOTE: This should not be used an accurate benchmark of Python code, but a
    way to check how much time has elapsed between actions. And this does not
    account for changes or blips in the system clock.
    Instance attributes:
    start_time -- timestamp when the timer started
    stop_time -- timestamp when the timer stopped
    As a regular object:
    >>> stopwatch = Stopwatch()
    >>> stopwatch.start()
    >>> time.sleep(1)
    >>> 1 <= stopwatch.time_elapsed <= 2
    True
    >>> time.sleep(1)
    >>> stopwatch.stop()
    >>> 2 <= stopwatch.total_run_time
    True
    As a context manager:
    >>> with Stopwatch() as stopwatch:
    ...     time.sleep(1)
    ...     print repr(1 <= stopwatch.time_elapsed <= 2)
    ...     time.sleep(1)
    True
    >>> 2 <= stopwatch.total_run_time
    True
    """

    def __init__(self):
        """Initialize a new `Stopwatch`, but do not start timing."""
        self.start_time = None
        self.stop_time = None

    def start(self):
        """Start timing."""
        self.start_time = time.time()

    def stop(self):
        """Stop timing."""
        self.stop_time = time.time()

    @property
    def time_elapsed(self):
        """Return the number of seconds that have elapsed since this
        `Stopwatch` started timing.
        This is used for checking how much time has elapsed while the timer is
        still running.
        """
        assert not self.stop_time, "Can't check `time_elapsed` on an ended `Stopwatch`."
        return time.time() - self.start_time

    @property
    def total_run_time(self):
        """Return the number of seconds that elapsed from when this `Stopwatch`
        started to when it ended.
        """
        return self.stop_time - self.start_time

    def __enter__(self):
        """Start timing and return this `Stopwatch` instance."""
        self.start()
        return self

    def __exit__(self, type, value, traceback):
        """Stop timing.
        If there was an exception inside the `with` block, re-raise it.
        >>> with Stopwatch() as stopwatch:
        ...     raise Exception
        Traceback (most recent call last):
            ...
        Exception
        """
        self.stop()


class DocumentRNNEmbeddingsFast(DocumentEmbeddings):
    def __init__(
        self,
        embeddings: List[TokenEmbeddings],
        hidden_size=128,
        rnn_layers=1,
        reproject_words: bool = True,
        reproject_words_dimension: int = None,
        bidirectional: bool = False,
        dropout: float = 0.5,
        word_dropout: float = 0.0,
        locked_dropout: float = 0.0,
        rnn_type="GRU",
    ):
        """The constructor takes a list of embeddings to be combined.
        :param embeddings: a list of token embeddings
        :param hidden_size: the number of hidden states in the rnn
        :param rnn_layers: the number of layers for the rnn
        :param reproject_words: boolean value, indicating whether to reproject the token embeddings in a separate linear
        layer before putting them into the rnn or not
        :param reproject_words_dimension: output dimension of reprojecting token embeddings. If None the same output
        dimension as before will be taken.
        :param bidirectional: boolean value, indicating whether to use a bidirectional rnn or not
        :param dropout: the dropout value to be used
        :param word_dropout: the word dropout value to be used, if 0.0 word dropout is not used
        :param locked_dropout: the locked dropout value to be used, if 0.0 locked dropout is not used
        :param rnn_type: 'GRU' or 'LSTM'
        """
        super().__init__()

        self.embeddings: StackedEmbeddings = StackedEmbeddings(embeddings=embeddings)

        self.rnn_type = rnn_type

        self.reproject_words = reproject_words
        self.bidirectional = bidirectional

        self.length_of_all_token_embeddings: int = self.embeddings.embedding_length

        self.static_embeddings = False

        self.__embedding_length: int = hidden_size
        if self.bidirectional:
            self.__embedding_length *= 4

        self.embeddings_dimension: int = self.length_of_all_token_embeddings
        if self.reproject_words and reproject_words_dimension is not None:
            self.embeddings_dimension = reproject_words_dimension

        self.word_reprojection_map = torch.nn.Linear(self.length_of_all_token_embeddings, self.embeddings_dimension)

        # bidirectional RNN on top of embedding layer
        if rnn_type == "LSTM":
            self.rnn = torch.nn.LSTM(
                self.embeddings_dimension,
                hidden_size,
                num_layers=rnn_layers,
                bidirectional=self.bidirectional,
            )
        else:
            self.rnn = torch.nn.GRU(
                self.embeddings_dimension,
                hidden_size,
                num_layers=rnn_layers,
                bidirectional=self.bidirectional,
            )

        self.name = "document_" + self.rnn._get_name()

        # dropouts
        if locked_dropout > 0.0:
            self.dropout: torch.nn.Module = LockedDropout(locked_dropout)
        else:
            self.dropout = torch.nn.Dropout(dropout)

        self.use_word_dropout: bool = word_dropout > 0.0
        if self.use_word_dropout:
            self.word_dropout = WordDropout(word_dropout)

        torch.nn.init.xavier_uniform_(self.word_reprojection_map.weight)

        self.to(flair.device)

    @property
    def embedding_length(self) -> int:
        return self.__embedding_length

    def embed(self, sentences: Union[List[Sentence], Sentence]):
        """Add embeddings to all sentences in the given list of sentences. If embeddings are already added, update
        only if embeddings are non-static."""

        if type(sentences) is Sentence:
            sentences = [sentences]

        self.rnn.zero_grad()

        sentences.sort(key=lambda x: len(x), reverse=True)

        self.embeddings.embed(sentences)

        # first, sort sentences by number of tokens
        longest_token_sequence_in_batch: int = len(sentences[0])

        all_sentence_tensors = []
        lengths: List[int] = []

        # go through each sentence in batch
        for i, sentence in enumerate(sentences):
            lengths.append(len(sentence.tokens))

            sentence_tensor = torch.cat([token.get_embedding().unsqueeze(0) for token in sentence], 0).to(flair.device)

            # ADD TO SENTENCE LIST: add the representation
            all_sentence_tensors.append(sentence_tensor)

        # --------------------------------------------------------------------
        # FF PART
        # --------------------------------------------------------------------
        # use word dropout if set
        # if self.use_word_dropout:
        #     sentence_tensor = self.word_dropout(sentence_tensor)

        # if self.reproject_words:
        #     sentence_tensor = self.word_reprojection_map(sentence_tensor)

        # sentence_tensor = self.dropout(sentence_tensor)

        packed = torch.nn.utils.rnn.pack_sequence(all_sentence_tensors, lengths)

        self.rnn.flatten_parameters()

        rnn_out, hidden = self.rnn(packed)

        outputs, output_lengths = torch.nn.utils.rnn.pad_packed_sequence(rnn_out)

        outputs = self.dropout(outputs)

        # --------------------------------------------------------------------
        # EXTRACT EMBEDDINGS FROM RNN
        # --------------------------------------------------------------------
        for sentence_no, length in enumerate(lengths):
            last_rep = outputs[length - 1, sentence_no]

            embedding = last_rep
            if self.bidirectional:
                first_rep = outputs[0, sentence_no]
                embedding = torch.cat([first_rep, last_rep], 0)

            sentence = sentences[sentence_no]
            sentence.set_embedding(self.name, embedding)

    def _add_embeddings_internal(self, sentences: List[Sentence]):
        pass


class MemoryEmbeddingsFast(TokenEmbeddings):
    def __init__(
        self,
        contextual_embeddings: TokenEmbeddings,
        concat_word_embeddings: bool = True,
        hidden_states: int = 64,
        reproject: bool = True,
        max_memory_length: int = 8,
        dropout: float = 0.5,
        **kwargs,
    ):

        super().__init__()

        # variables
        self.hidden_states = hidden_states
        self.max_memory_length: int = max_memory_length
        self.effective_memory_length: int = self.max_memory_length
        self.concat_word_embeddings: bool = concat_word_embeddings
        self.static_embeddings: bool = False

        # use the character language model embeddings as basis
        if type(contextual_embeddings) is str:
            self.context_embeddings: FlairEmbeddings = FlairEmbeddings(contextual_embeddings, **kwargs)
        else:
            self.context_embeddings: TokenEmbeddings = contextual_embeddings

        self.sub_embedding_names = (
            [emb.name for emb in self.context_embeddings.embeddings]
            if type(self.context_embeddings) is StackedEmbeddings
            else [self.context_embeddings.name]
        )

        # determine embedding length
        self.__embedding_length = (
            self.context_embeddings.embedding_length + self.hidden_states
            if self.concat_word_embeddings
            else 0 + self.hidden_states
        )

        # the memory
        self.word_history = {}
        self.state_history = {}

        # the NN
        self.dropout = torch.nn.Dropout(dropout)

        self.reproject: bool = reproject
        if self.reproject:
            self.reprojection_layer = torch.nn.Linear(
                self.context_embeddings.embedding_length,
                self.context_embeddings.embedding_length,
            )

        self.rnn = torch.nn.GRU(
            self.context_embeddings.embedding_length,
            self.hidden_states,
            num_layers=1,
            bidirectional=False,
        )
        self.name = self.context_embeddings.name + "-memory"
        #
        self.all_time = 0.0
        self.prepare_time = 0.0
        self.nn_time = 0.0
        self.add_to_memory_time = 0.0
        self.add_to_words_time = 0.0
        self.pad_time = 0.0
        self.rest_time = 0.0
        self.batch_count = 0

        self.to(flair.device)

    def train(self, mode=True):
        super().train(mode=mode)
        compress = False
        if mode:
            # memory is wiped each time we do a training run
            print("train mode resetting embeddings")
            self.word_history = {}
            self.state_history = {}
            print(self.word_history)
            print(self.state_history)
            self.effective_memory_length = self.max_memory_length
            print(self.effective_memory_length)
        elif compress:
            # memory is wiped each time we do a training run
            print("prediction mode no backprop")
            self.word_history = {}
            for word in self.state_history:
                self.state_history[word] = [self.state_history[word][-1].clone()]
            self.effective_memory_length = 1

    def _add_embeddings_internal(self, sentences: List[Sentence]) -> List[Sentence]:

        self.context_embeddings.embed(sentences)
        #
        # all_stopwatch = Stopwatch()
        # prepare_stopwatch = Stopwatch()
        # nn_stopwatch = Stopwatch()
        # add_to_memory_stopwatch = Stopwatch()
        # add_to_words_stopwatch = Stopwatch()
        # pad_stopwatch = Stopwatch()
        # rest_stopwatch = Stopwatch()
        #
        # all_stopwatch.start()

        # padding = torch.zeros(
        #     self.context_embeddings.embedding_length, dtype=torch.float
        # ).to(flair.device)

        # determine and add to history of each token
        surface_form_history = {}
        for sentence in sentences:
            for token in sentence:
                # add current embedding to the memory
                local_embedding = token.get_embedding(self.sub_embedding_names)

                if token.text not in self.word_history:
                    self.word_history[token.text] = [local_embedding]
                else:
                    self.word_history[token.text].append(local_embedding)

                surface_form_history[token.text] = self.word_history[token.text]

        # sort surface forms by longest history
        surface_forms_sorted_by_memory_length = sorted(
            surface_form_history,
            key=lambda k: len(surface_form_history[k]),
            reverse=True,
        )
        logest_memory_length_in_batch = len(surface_form_history[surface_forms_sorted_by_memory_length[0]])

        # all_surface_form_histories = []

        # initialize zero-padded word embeddings tensor
        all_surface_form_tensor = torch.zeros(
            [
                logest_memory_length_in_batch,
                len(surface_form_history),
                self.context_embeddings.embedding_length,
            ],
            dtype=torch.float,
            device=flair.device,
        )

        # default zero-state tensor
        zeros = torch.zeros(self.hidden_states, device=flair.device)

        all_surface_form_initial_hidden = []
        lengths = []

        # prepare_stopwatch.start()

        # print('all:')
        # print(all_surface_form_tensor)
        # go through each unique surface form
        for s_id, surface_form in enumerate(surface_forms_sorted_by_memory_length):

            # pad_stopwatch.start()

            # get embedding history of this surface form and bring to flair.device
            surface_form_embedding_history = surface_form_history[surface_form]

            # for i in range(len(surface_form_embedding_history)):
            #     surface_form_embedding_history[i] = surface_form_embedding_history[i].to(flair.device)

            # pad shorter sequences out
            # for add in range(logest_memory_length_in_batch - len(surface_form_embedding_history)):
            #     surface_form_embedding_history.append(padding)

            for i in range(len(surface_form_embedding_history)):
                # print('histry:')
                # print(surface_form_embedding_history[i])
                all_surface_form_tensor[i, s_id] = surface_form_embedding_history[i]

            # print('all:')
            # print(all_surface_form_tensor)
            # all_surface_form_tensor[s_id][: len(sentence)] = torch.stack(
            #     [token.get_embedding() for token in sentence], 0
            # )

            lengths.append(len(surface_form_embedding_history))

            # pad_stopwatch.stop()
            # self.pad_time += pad_stopwatch.total_run_time
            # rest_stopwatch.start()

            # truncate surface form history if necessary
            window = (
                0
                if len(self.word_history[surface_form]) < self.effective_memory_length
                else len(self.word_history[surface_form]) - self.effective_memory_length
            )
            if window > 0:
                self.word_history[surface_form] = self.word_history[surface_form][
                    window : self.effective_memory_length + window
                ]

            # print(surface_form_embedding_history)

            # make padded tensor of history
            # surface_form_embedding_history = torch.stack(surface_form_embedding_history, dim=0)

            # surface_form_embedding_history =  torch.cat(surface_form_embedding_history).view(-1, self.context_embeddings.embedding_length)

            # surface_form_embedding_history = surface_form_embedding_history.to(flair.device)

            # add history to all
            # all_surface_form_histories.append(surface_form_embedding_history
            # )

            # initialize first hidden state if necessary
            if surface_form not in self.state_history:
                self.state_history[surface_form] = [zeros]

            all_surface_form_initial_hidden.append(self.state_history[surface_form][0])

            # rest_stopwatch.stop()
            # self.rest_time += rest_stopwatch.total_run_time

        # make batch tensors
        # all_surface_form_histories = torch.stack(all_surface_form_histories, 1)
        all_surface_form_histories = all_surface_form_tensor
        # print(all_surface_form_tensor)
        # asd

        all_surface_form_initial_hidden = torch.stack(all_surface_form_initial_hidden, 0).unsqueeze(0)

        # prepare_stopwatch.stop()
        #
        # nn_stopwatch.start()

        # dropout!
        all_surface_form_histories = self.dropout(all_surface_form_histories)
        # print(all_surface_form_histories)

        # reproject if set
        if self.reproject:
            all_surface_form_histories = self.reprojection_layer(all_surface_form_histories)

        # send through RNN
        # rnn_out, hidden = self.rnn(all_surface_form_histories, all_surface_form_initial_hidden)
        packed = torch.nn.utils.rnn.pack_padded_sequence(all_surface_form_histories, lengths)

        packed_output, hidden = self.rnn(packed, all_surface_form_initial_hidden)

        rnn_out, hidden = torch.nn.utils.rnn.pad_packed_sequence(packed_output)

        # nn_stopwatch.stop()
        #
        # add_to_memory_stopwatch.start()

        # go through each unique surface form and update state history
        for idx, surface_form in enumerate(surface_forms_sorted_by_memory_length):
            # retrieve hidden states for this surface form, and the last initial hidden state
            hidden_states_of_surface_form = rnn_out[0 : lengths[idx], idx]
            last_initial_hidden_state = all_surface_form_initial_hidden[0][idx].unsqueeze(0)

            # concat both to get new hidden state history
            hidden_state_history = torch.cat([last_initial_hidden_state, hidden_states_of_surface_form], dim=0)

            # if history is too long, truncate to window
            window = (
                0
                if hidden_state_history.size(0) < self.effective_memory_length
                else hidden_state_history.size(0) - self.effective_memory_length
            )
            state_history_window = hidden_state_history.detach()[window : self.effective_memory_length + window]

            # set as new state history of surface form (before: .cpu())
            self.state_history[surface_form] = state_history_window

        # add_to_memory_stopwatch.stop()
        #
        # add_to_words_stopwatch.start()

        # finally, go through each token of each sentence and set the embedding
        for sentence in sentences:

            for token in sentence:

                idx = surface_forms_sorted_by_memory_length.index(token.text)

                embedding = rnn_out[lengths[idx] - 1, idx]

                # print(self.state_history[token.text][-1])

                token.set_embedding(self.name, embedding)

                if not self.concat_word_embeddings:
                    for subembedding in self.sub_embedding_names:
                        if subembedding in token._embeddings.keys():
                            del token._embeddings[subembedding]

        # add_to_words_stopwatch.stop()
        # all_stopwatch.stop()
        # self.all_time += all_stopwatch.total_run_time
        # self.prepare_time += prepare_stopwatch.total_run_time
        # self.nn_time += nn_stopwatch.total_run_time
        # self.add_to_memory_time += add_to_memory_stopwatch.total_run_time
        # self.add_to_words_time += add_to_words_stopwatch.total_run_time
        # self.batch_count += 1
        # print(f'Total time    : {self.all_time}')
        # print(f' - Preparation: {round(self.prepare_time / self.all_time, 2)}% ({self.prepare_time})')
        # print(f'      - Pad   : {round(self.pad_time / self.prepare_time, 2)}% ({self.pad_time})')
        # print(f'      - Rest  : {round(self.rest_time / self.prepare_time, 2)}% ({self.rest_time})')
        # print(f' - NN + RNN   : {round(self.nn_time / self.all_time, 2)}% ({self.nn_time})')
        # print(f' - Memory     : {round(self.add_to_memory_time / self.all_time, 2)}% ({self.add_to_memory_time})')
        # print(f' - Add Words  : {round(self.add_to_words_time / self.all_time, 2)}% ({self.add_to_words_time})')
        # print(f'Batches: {self.batch_count}')
        # print()
        #
        # if self.all_time > 10:
        #     asd

        return sentences

    @property
    def embedding_length(self) -> int:
        return self.__embedding_length

    def get_names(self) -> List[str]:
        """Returns a list of embedding names. In most cases, it is just a list with one item, namely the name of
        this embedding. But in some cases, the embedding is made up by different embeddings (StackedEmbedding).
        Then, the list contains the names of all embeddings in the stack."""
        names = [self.name]
        names.extend(self.context_embeddings.get_names())
        return names


class MemoryEmbeddings(TokenEmbeddings):
    def __init__(
        self,
        contextual_embeddings: TokenEmbeddings,
        concat_word_embeddings: bool = True,
        hidden_states: int = 64,
        reproject: bool = True,
        max_memory_length: int = 8,
        dropout: float = 0.5,
        use_batch_memory_state: bool = False,
    ):

        super().__init__()

        # variables
        self.hidden_states = hidden_states
        self.max_memory_length: int = max_memory_length
        self.effective_memory_length: int = self.max_memory_length
        self.concat_word_embeddings: bool = concat_word_embeddings
        self.static_embeddings: bool = False

        self.context_embeddings: TokenEmbeddings = contextual_embeddings

        self.sub_embedding_names = (
            [emb.name for emb in self.context_embeddings.embeddings]
            if type(self.context_embeddings) is StackedEmbeddings
            else [self.context_embeddings.name]
        )

        # determine embedding length
        self.__embedding_length = (
            self.context_embeddings.embedding_length + self.hidden_states
            if self.concat_word_embeddings
            else 0 + self.hidden_states
        )

        # the memory
        self.word_history = {}
        self.state_history = {}

        # the NN
        self.dropout = torch.nn.Dropout(dropout)

        self.reproject: bool = reproject
        if self.reproject:
            self.reprojection_layer = torch.nn.Linear(
                self.context_embeddings.embedding_length,
                self.context_embeddings.embedding_length,
            )

        self.rnn = torch.nn.GRU(
            self.context_embeddings.embedding_length,
            self.hidden_states,
            num_layers=1,
            bidirectional=False,
        )
        self.name = self.context_embeddings.name + "-memory"

        self.use_batch_memory_state = use_batch_memory_state

        self.to(flair.device)

    def train(self, mode=True):
        super().train(mode=mode)
        compress = False
        if mode:
            # memory is wiped each time we do a training run
            print("train mode resetting embeddings")
            self.word_history = {}
            self.state_history = {}
            print(self.word_history)
            print(self.state_history)
            self.effective_memory_length = self.max_memory_length
            print(self.effective_memory_length)
        elif compress:
            # memory is wiped each time we do a training run
            print("prediction mode no backprop")
            self.word_history = {}
            for word in self.state_history:
                self.state_history[word] = [self.state_history[word][-1].clone()]
            self.effective_memory_length = 1

    def _add_embeddings_internal(self, sentences: List[Sentence]) -> List[Sentence]:

        self.context_embeddings.embed(sentences)

        # determine and add to history of each token
        surface_form_history = {}
        for sentence in sentences:
            for token in sentence:
                # add current embedding to the memory
                local_embedding = token.get_embedding(self.sub_embedding_names)

                if token.text not in self.word_history:
                    self.word_history[token.text] = [local_embedding]
                else:
                    self.word_history[token.text].append(local_embedding)

                surface_form_history[token.text] = self.word_history[token.text]

        # sort surface forms by longest history
        surface_forms_sorted_by_memory_length = sorted(
            surface_form_history,
            key=lambda k: len(surface_form_history[k]),
            reverse=True,
        )
        logest_memory_length_in_batch = len(surface_form_history[surface_forms_sorted_by_memory_length[0]])
        # print(surface_forms_sorted_by_memory_length)

        # initialize zero-padded word embeddings tensor
        all_surface_form_tensor = torch.zeros(
            [
                logest_memory_length_in_batch,
                len(surface_form_history),
                self.context_embeddings.embedding_length,
            ],
            dtype=torch.float,
            device=flair.device,
        )
        # default zero-state tensor
        zeros = torch.zeros(self.hidden_states, device=flair.device)

        all_surface_form_initial_hidden = []
        lengths = []

        # go through each unique surface form
        for s_id, surface_form in enumerate(surface_forms_sorted_by_memory_length):

            # get embedding history of this surface form and bring to flair.device
            surface_form_embedding_history = surface_form_history[surface_form]
            for i in range(len(surface_form_embedding_history)):
                # TODO: use dropout only on first one
                all_surface_form_tensor[i, s_id] = surface_form_embedding_history[i]

            lengths.append(len(surface_form_embedding_history))

            # truncate surface form history if necessary
            window = (
                0
                if len(self.word_history[surface_form]) < self.effective_memory_length
                else len(self.word_history[surface_form]) - self.effective_memory_length
            )
            if window > 0:
                self.word_history[surface_form] = self.word_history[surface_form][
                    window : self.effective_memory_length + window
                ]

            # initialize first hidden state if necessary
            if surface_form not in self.state_history:
                self.state_history[surface_form] = zeros

            all_surface_form_initial_hidden.append(self.state_history[surface_form])

        # make batch tensors
        all_surface_form_histories = all_surface_form_tensor

        all_surface_form_initial_hidden = torch.stack(all_surface_form_initial_hidden, 0).unsqueeze(0)

        # print(all_surface_form_initial_hidden)
        # print(all_surface_form_initial_hidden.size())

        # dropout!
        all_surface_form_histories = self.dropout(all_surface_form_histories)

        # reproject if set
        if self.reproject:
            all_surface_form_histories = self.reprojection_layer(all_surface_form_histories)

        # send through RNN
        packed = torch.nn.utils.rnn.pack_padded_sequence(all_surface_form_histories, lengths)

        packed_output, hidden = self.rnn(packed, all_surface_form_initial_hidden)

        rnn_out, hidden = torch.nn.utils.rnn.pad_packed_sequence(packed_output)

        # print('rnn out: ')
        # print(rnn_out.size())
        # print(rnn_out)

        if not self.training:
            # go through each unique surface form and update state history
            for idx, surface_form in enumerate(surface_forms_sorted_by_memory_length):
                # retrieve hidden states for this surface form, and the last initial hidden state
                self.state_history[surface_form] = rnn_out[lengths[idx] - 1, idx]

        # print(self.state_history)

        # finally, go through each token of each sentence and set the embedding
        surface_form_counter = Counter()
        for sentence in sentences:

            for token in sentence:
                surface_form_counter[token.text] += 1

                idx = surface_forms_sorted_by_memory_length.index(token.text)
                if self.use_batch_memory_state:
                    memory_index = lengths[idx] - 1
                else:
                    memory_index = surface_form_counter[token.text] - 1

                embedding = rnn_out[memory_index, idx]

                if self.concat_word_embeddings:
                    import random

                    stack = False if self.training and random.randint(1, 100) <= 50 else True

                    embedding = (
                        torch.cat([embedding, token.get_embedding(self.sub_embedding_names)])
                        if stack
                        else torch.cat(
                            [
                                embedding,
                                torch.zeros(
                                    self.embedding_length - self.hidden_states,
                                    device=flair.device,
                                ),
                            ]
                        )
                    )

                token.set_embedding(self.name, embedding)

        return sentences

    @property
    def embedding_length(self) -> int:
        return self.__embedding_length

    def extra_repr(self):
        return f"[use_batch_memory_state='{self.use_batch_memory_state}', concat_word_embeddings='{self.concat_word_embeddings}']"

    # def get_names(self) -> List[str]:
    #     """Returns a list of embedding names. In most cases, it is just a list with one item, namely the name of
    #     this embedding. But in some cases, the embedding is made up by different embeddings (StackedEmbedding).
    #     Then, the list contains the names of all embeddings in the stack."""
    #     names = [self.name]
    #     names.extend(self.context_embeddings.get_names())
    #     return names


class OrderedMemoryEmbeddings(TokenEmbeddings):
    def __init__(
        self,
        contextual_embeddings: TokenEmbeddings,
        concat_word_embeddings: bool = True,
        hidden_states: int = 64,
        reproject: bool = False,
        max_memory_length: int = 8,
        dropout: float = 0.0,
        use_batch_memory_state: bool = False,
        train_initial_hidden_state: bool = True,
    ):

        super().__init__()

        # variables
        self.hidden_states = hidden_states
        self.max_memory_length: int = max_memory_length
        self.effective_memory_length: int = self.max_memory_length
        self.concat_word_embeddings: bool = concat_word_embeddings
        self.static_embeddings: bool = False
        self.train_initial_hidden_state = train_initial_hidden_state

        self.context_embeddings: TokenEmbeddings = contextual_embeddings

        self.sub_embedding_names = (
            [emb.name for emb in self.context_embeddings.embeddings]
            if type(self.context_embeddings) is StackedEmbeddings
            else [self.context_embeddings.name]
        )

        # determine embedding length
        self.__embedding_length = (
            self.context_embeddings.embedding_length + self.hidden_states
            if self.concat_word_embeddings
            else 0 + self.hidden_states
        )

        # the memory
        self.last_instance = {}
        self.last_state = {}

        # the NN
        self.dropout = torch.nn.Dropout(dropout)

        self.reproject: bool = reproject
        if self.reproject:
            self.reprojection_layer = torch.nn.Linear(
                self.context_embeddings.embedding_length,
                self.context_embeddings.embedding_length,
            )

        self.rnn = torch.nn.GRU(
            self.context_embeddings.embedding_length,
            self.hidden_states,
            num_layers=1,
            bidirectional=False,
        )
        if self.train_initial_hidden_state:
            self.initial_hidden = Parameter(torch.zeros(self.hidden_states), requires_grad=True)
        else:
            self.initial_hidden = torch.zeros(self.hidden_states)

        self.name = self.context_embeddings.name + "+memory"

        self.use_batch_memory_state = use_batch_memory_state

        self.to(flair.device)

    def train(self, mode=True):
        super().train(mode=mode)

        if mode:
            # memory is wiped each time we do a training run
            print("train mode resetting embeddings")
            self.last_state = {}
            self.effective_memory_length = self.max_memory_length
            print(self.effective_memory_length)

        else:
            # memory is wiped each time we do a training run
            print("prediction mode no backprop")
            for word in self.last_instance.keys():
                self.last_state[word] = getattr(self.last_instance[word], "last_hidden")
            self.effective_memory_length = 0

            # print(self.last_state)

    def _add_embeddings_internal(self, sentences: List[Sentence]) -> List[Sentence]:

        word_history = {}

        self.context_embeddings.embed(sentences)

        # default zero-state tensor
        # zeros = torch.zeros(self.hidden_states, device=flair.device)

        # go through all tokens and set previous instance history if not yet set

        if self.training:
            idx = 0
            for sentence in sentences:
                for token in sentence:

                    if not hasattr(token, "previous_instance"):
                        if token.text in self.last_instance.keys():
                            setattr(
                                token,
                                "previous_instance",
                                self.last_instance[token.text],
                            )
                        else:
                            setattr(token, "previous_instance", None)
                    idx += 1

                    self.last_instance[token.text] = token
        else:
            idx = 0
            for sentence in sentences:
                for token in sentence:
                    idx += 1

        # initialize zero-padded word embeddings tensor
        all_surface_form_tensor = torch.zeros(
            [
                self.effective_memory_length + 1,
                idx,
                self.context_embeddings.embedding_length,
            ],
            dtype=torch.float,
            device=flair.device,
        )

        all_hidden_tensor = torch.zeros(
            [
                1,
                idx,
                self.hidden_states,
            ],
            dtype=torch.float,
            device=flair.device,
        )

        # print(all_surface_form_tensor.size())

        idx = 0
        lengths = []
        for sentence in sentences:
            for token in sentence:

                word_history[idx] = [token]
                memory = token
                if self.training:
                    while True:
                        memory = memory.previous_instance
                        if memory == None:
                            break
                        if len(word_history[idx]) >= self.effective_memory_length + 1:
                            break

                        word_history[idx].insert(0, memory)

                for m_id, memory_token in enumerate(word_history[idx]):
                    all_surface_form_tensor[m_id, idx] = memory_token.get_embedding(self.sub_embedding_names)

                all_hidden_tensor[0, idx] = (
                    self.initial_hidden if token.text not in self.last_state.keys() else self.last_state[token.text]
                )

                lengths.append(len(word_history[idx]) - 1)  # length of history only (do not count current token)

                idx += 1

        # make batch tensors
        all_surface_form_histories = all_surface_form_tensor

        # dropout!
        all_surface_form_histories = self.dropout(all_surface_form_histories)

        # reproject if set
        if self.reproject:
            all_surface_form_histories = self.reprojection_layer(all_surface_form_histories)

        # print(all_surface_form_histories.size())
        rnn_out, hidden = self.rnn(all_surface_form_histories, all_hidden_tensor)
        # print(rnn_out.size())

        # finally, go through each token of each sentence and set the embedding
        idx = 0
        for sentence in sentences:

            for token in sentence:

                last_memory_index = lengths[idx] - 1
                updated_memory_index = lengths[idx]

                if last_memory_index >= 0:
                    embedding = rnn_out[last_memory_index, idx]
                else:
                    embedding = (
                        self.initial_hidden if token.text not in self.last_state.keys() else self.last_state[token.text]
                    )

                # remember last hidden state of token
                last_hidden = rnn_out[updated_memory_index, idx]

                if self.training:
                    setattr(token, "last_hidden", last_hidden.clone().detach())
                else:
                    self.last_state[token.text] = last_hidden.clone().detach()

                if self.concat_word_embeddings:
                    import random

                    stack = False if self.training and random.randint(1, 100) <= 50 else True

                    embedding = (
                        torch.cat([embedding, token.get_embedding(self.sub_embedding_names)])
                        if stack
                        else torch.cat(
                            [
                                embedding,
                                torch.zeros(
                                    self.embedding_length - self.hidden_states,
                                    device=flair.device,
                                ),
                            ]
                        )
                    )

                token.set_embedding(self.name, embedding)

                idx += 1

        return sentences

    @property
    def embedding_length(self) -> int:
        return self.__embedding_length

    def extra_repr(self):
        return f"[use_batch_memory_state='{self.use_batch_memory_state}', concat_word_embeddings='{self.concat_word_embeddings}']"

    def __getstate__(self):

        for word in self.last_instance.keys():
            self.last_state[word] = getattr(self.last_instance[word], "last_hidden")
        self.last_instance = {}

        return self.__dict__



import logging
from typing import List, Union

import torch
import torch.nn

import flair.nn
from flair.data import Dictionary, Sentence
from flair.embeddings import TokenEmbeddings

log = logging.getLogger("flair")


class WordTagger(flair.nn.DefaultClassifier[Sentence]):
    """
    This is a simple class of models that tags individual words in text.
    """

    def __init__(
        self,
        embeddings: TokenEmbeddings,
        tag_dictionary: Dictionary,
        tag_type: str,
        **classifierargs,
    ):
        """
        Initializes a WordTagger
        :param embeddings: word embeddings used in tagger
        :param tag_dictionary: dictionary of tags you want to predict
        :param tag_type: string identifier for tag type
        :param beta: Parameter for F-beta score for evaluation and training annealing
        """
        super().__init__(
            label_dictionary=tag_dictionary, final_embedding_size=embeddings.embedding_length, **classifierargs
        )

        # embeddings
        self.embeddings = embeddings

        # dictionaries
        self.tag_type: str = tag_type

        # all parameters will be pushed internally to the specified device
        self.to(flair.device)

    def _get_state_dict(self):
        model_state = {
            **super()._get_state_dict(),
            "embeddings": self.embeddings,
            "tag_dictionary": self.label_dictionary,
            "tag_type": self.tag_type,
        }
        return model_state

    @classmethod
    def _init_model_with_state_dict(cls, state, **kwargs):
        return super()._init_model_with_state_dict(
            state,
            embeddings=state["embeddings"],
            tag_dictionary=state["tag_dictionary"],
            tag_type=state["tag_type"],
            **kwargs,
        )

    def forward_pass(
        self,
        sentences: Union[List[Sentence], Sentence],
        for_prediction: bool = False,
    ):
        if not isinstance(sentences, list):
            sentences = [sentences]

        self.embeddings.embed(sentences)

        names = self.embeddings.get_names()

        # get all tokens in this mini-batch
        all_tokens = [token for sentence in sentences for token in sentence]

        all_embeddings = [token.get_embedding(names) for token in all_tokens]

        embedded_tokens = torch.stack(all_embeddings)

        labels = [[token.get_label(self.label_type).value] for token in all_tokens]

        if for_prediction:
            return embedded_tokens, labels, all_tokens

        return embedded_tokens, labels

    @property
    def label_type(self):
        return self.tag_type

    def _print_predictions(self, batch, gold_label_type):
        lines = []
        for datapoint in batch:
            # now print labels in CoNLL format
            for token in datapoint:
                eval_line = (
                    f"{token.text} "
                    f"{token.get_label(gold_label_type, 'O').value} "
                    f"{token.get_label('predicted', 'O').value}\n"
                )
                lines.append(eval_line)
            lines.append("\n")
        return lines



from .clustering import ClusteringModel
from .dependency_parser_model import DependencyParser
from .entity_linker_model import EntityLinker
from .language_model import LanguageModel
from .lemmatizer_model import Lemmatizer
from .pairwise_classification_model import TextPairClassifier
from .regexp_tagger import RegexpTagger
from .relation_extractor_model import RelationExtractor
from .sequence_tagger_model import MultiTagger, SequenceTagger
from .tars_model import FewshotClassifier, TARSClassifier, TARSTagger
from .text_classification_model import TextClassifier
from .word_tagger_model import WordTagger

__all__ = [
    "DependencyParser",
    "EntityLinker",
    "LanguageModel",
    "Lemmatizer",
    "TextPairClassifier",
    "RelationExtractor",
    "RegexpTagger",
    "MultiTagger",
    "SequenceTagger",
    "WordTagger",
    "FewshotClassifier",
    "TARSClassifier",
    "TARSTagger",
    "TextClassifier",
    "ClusteringModel",
]



import torch

# Source: https://github.com/UKPLab/sentence-transformers/blob/master/sentence_transformers/util.py#L23


def dot_product(a: torch.Tensor, b: torch.Tensor, normalize=False):
    """
    Computes dot product for pairs of vectors.
    :param normalize: Vectors are normalized (leads to cosine similarity)
    :return: Matrix with res[i][j]  = dot_product(a[i], b[j])
    """
    if len(a.shape) == 1:
        a = a.unsqueeze(0)

    if len(b.shape) == 1:
        b = b.unsqueeze(0)

    if normalize:
        a = torch.nn.functional.normalize(a, p=2, dim=1)
        b = torch.nn.functional.normalize(b, p=2, dim=1)

    return torch.mm(a, b.transpose(0, 1))


class CosineDistance(torch.nn.Module):
    def forward(self, a, b):
        return -dot_product(a, b, normalize=True)


class LogitCosineDistance(torch.nn.Module):
    def forward(self, a, b):
        return torch.logit(0.5 - 0.5 * dot_product(a, b, normalize=True))


class NegativeScaledDotProduct(torch.nn.Module):
    def forward(self, a, b):
        sqrt_d = torch.sqrt(torch.tensor(a.size(-1)))
        return -dot_product(a, b, normalize=False) / sqrt_d



from .param_selection import (
    SearchSpace,
    SequenceTaggerParamSelector,
    TextClassifierParamSelector,
)
from .parameter import (
    SEQUENCE_TAGGER_PARAMETERS,
    TEXT_CLASSIFICATION_PARAMETERS,
    TRAINING_PARAMETERS,
    Parameter,
)

__all__ = [
    "Parameter",
    "SEQUENCE_TAGGER_PARAMETERS",
    "TRAINING_PARAMETERS",
    "TEXT_CLASSIFICATION_PARAMETERS",
    "SequenceTaggerParamSelector",
    "TextClassifierParamSelector",
    "SearchSpace",
]



"""
This module was copied from the repository the following repository:
https://github.com/asappresearch/dynamic-classification

It contains the code from the paper "Metric Learning for Dynamic Text
Classification".

https://arxiv.org/abs/1911.01026

In case this file is modified, please consider contributing to the original
repository.

It was published under MIT License:
https://github.com/asappresearch/dynamic-classification/blob/master/LICENSE.md

Source: https://github.com/asappresearch/dynamic-classification/blob/55beb5a48406c187674bea40487c011e8fa45aab/distance/euclidean.py
"""


import torch
import torch.nn as nn
from torch import Tensor


class EuclideanDistance(nn.Module):
    """Implement a EuclideanDistance object."""

    def forward(self, mat_1: Tensor, mat_2: Tensor) -> Tensor:  # type: ignore
        """Returns the squared euclidean distance between each
        element in mat_1 and each element in mat_2.

        Parameters
        ----------
        mat_1: torch.Tensor
            matrix of shape (n_1, n_features)
        mat_2: torch.Tensor
            matrix of shape (n_2, n_features)

        Returns
        -------
        dist: torch.Tensor
            distance matrix of shape (n_1, n_2)

        """
        _dist = [torch.sum((mat_1 - mat_2[i]) ** 2, dim=1) for i in range(mat_2.size(0))]
        dist = torch.stack(_dist, dim=1)
        return dist


class EuclideanMean(nn.Module):
    """Implement a EuclideanMean object."""

    def forward(self, data: Tensor) -> Tensor:  # type: ignore
        """Performs a forward pass through the network.

        Parameters
        ----------
        data : torch.Tensor
            The input data, as a float tensor

        Returns
        -------
        torch.Tensor
            The encoded output, as a float tensor

        """
        return data.mean(0)



import logging
import random
from collections import defaultdict
from typing import Dict

import torch
from torch.utils.data.sampler import Sampler

log = logging.getLogger("flair")


class FlairSampler(Sampler):
    def set_dataset(self, data_source):
        """Initialize by passing a block_size and a plus_window parameter.
        :param data_source: dataset to sample from
        """
        self.data_source = data_source
        self.num_samples = len(self.data_source)

    def __len__(self):
        return self.num_samples


class ImbalancedClassificationDatasetSampler(FlairSampler):
    """Use this to upsample rare classes and downsample common classes in your unbalanced classification dataset."""

    def __init__(self):
        super(ImbalancedClassificationDatasetSampler, self).__init__(None)

    def set_dataset(self, data_source):
        """
        Initialize by passing a classification dataset with labels, i.e. either TextClassificationDataSet or
        :param data_source:
        """
        self.data_source = data_source
        self.num_samples = len(self.data_source)
        self.indices = list(range(len(data_source)))

        # first determine the distribution of classes in the dataset
        label_count: Dict[str, int] = defaultdict(int)
        for sentence in data_source:
            for label in sentence.labels:
                label_count[label.value] += 1

        # weight for each sample
        offset = 0
        weights = [1.0 / (offset + label_count[data_source[idx].labels[0].value]) for idx in self.indices]

        self.weights = torch.DoubleTensor(weights)

    def __iter__(self):
        return (self.indices[i] for i in torch.multinomial(self.weights, self.num_samples, replacement=True))


class ChunkSampler(FlairSampler):
    """Splits data into blocks and randomizes them before sampling. This causes some order of the data to be preserved,
    while still shuffling the data.
    """

    def __init__(self, block_size=5, plus_window=5):
        super(ChunkSampler, self).__init__(None)
        self.block_size = block_size
        self.plus_window = plus_window
        self.data_source = None

    def __iter__(self):
        data = list(range(len(self.data_source)))

        blocksize = self.block_size + random.randint(0, self.plus_window)

        log.info(f"Chunk sampling with blocksize = {blocksize} ({self.block_size} + {self.plus_window})")

        # Create blocks
        blocks = [data[i : i + blocksize] for i in range(0, len(data), blocksize)]
        # shuffle the blocks
        random.shuffle(blocks)
        # concatenate the shuffled blocks
        data[:] = [b for bs in blocks for b in bs]
        return iter(data)


class ExpandingChunkSampler(FlairSampler):
    """Splits data into blocks and randomizes them before sampling. Block size grows with each epoch.
    This causes some order of the data to be preserved, while still shuffling the data.
    """

    def __init__(self, step=3):
        """Initialize by passing a block_size and a plus_window parameter.
        :param data_source: dataset to sample from
        """
        super(ExpandingChunkSampler, self).__init__(None)
        self.block_size = 1
        self.epoch_count = 0
        self.step = step

    def __iter__(self):
        self.epoch_count += 1

        data = list(range(len(self.data_source)))

        log.info(f"Chunk sampling with blocksize = {self.block_size}")

        # Create blocks
        blocks = [data[i : i + self.block_size] for i in range(0, len(data), self.block_size)]
        # shuffle the blocks
        random.shuffle(blocks)
        # concatenate the shuffled blocks
        data[:] = [b for bs in blocks for b in bs]

        if self.epoch_count % self.step == 0:
            self.block_size += 1

        return iter(data)
